@@ -11,15 +11,15 @@ use HTTP::Date;
 
 =head1 NAME
 
-CGI::Buffer - Verify and Optimise CGI Output
+CGI::Buffer - Verify, Cache and Optimise CGI Output
 
 =head1 VERSION
 
-Version 0.74
+Version 0.75
 
 =cut
 
-our $VERSION = '0.74';
+our $VERSION = '0.75';
 
 =head1 SYNOPSIS
 
@@ -83,7 +83,7 @@ our $logger;
 our $status;
 our $script_mtime;
 our $cobject;
-our($buf, $pos, $headers, $header, $body, @content_type, $etag,
+our($x_cache, $buf, $headers, $header, $body, @content_type, $etag,
 	$send_body, @o);
 
 BEGIN {
@@ -91,10 +91,29 @@ BEGIN {
 
 	$CGI::Buffer::buf = IO::String->new();
 	$CGI::Buffer::old_buf = select($CGI::Buffer::buf);
+
+	if((!defined($ENV{'SERVER_PROTOCOL'})) ||
+	  ($ENV{'SERVER_PROTOCOL'} eq 'HTTP/1.0')) {
+	  	$generate_etag = 0;
+	}
 }
 
 END {
+	if(defined($^V) && ($^V ge 'v5.14.0')) {
+		return if ${^GLOBAL_PHASE} eq 'DESTRUCT';	# >= 5.14.0 only
+	}
+
 	if($logger) {
+		if($ENV{'HTTP_IF_NONE_MATCH'}) {
+			$logger->debug("HTTP_IF_NONE_MATCH: $ENV{HTTP_IF_NONE_MATCH}");
+		}
+		if($ENV{'HTTP_IF_MODIFIED_SINCE'}) {
+			$logger->debug("HTTP_IF_MODIFIED_SINCE: $ENV{HTTP_IF_MODIFIED_SINCE}");
+		}
+		$logger->debug("Generate_etag = $generate_etag");
+		$logger->debug("Generate_304 = $generate_304");
+		$logger->debug("Generate_last_modified = $generate_last_modified");
+
 		# This will cause everything to get flushed and prevent
 		# outputs to the logger.  We need to do that now since
 		# if we leave it to Perl to delete later we may get
@@ -102,7 +121,7 @@ END {
 		$logger = undef;
 	}
 	select($CGI::Buffer::old_buf);
-	$pos = $CGI::Buffer::buf->getpos;
+	my $pos = $CGI::Buffer::buf->getpos;
 	$CGI::Buffer::buf->setpos(0);
 	read($CGI::Buffer::buf, $buf, $pos);
 	($headers, $body) = split /\r?\n\r?\n/, $buf, 2;
@@ -152,7 +171,7 @@ END {
 
 			# If we're on http://www.example.com and have a link
 			# to http://www.example.com/foo/bar.htm, change the
-			# link to /foo.bar.htm - there's no need to include
+			# link to /foo/bar.htm - there's no need to include
 			# the site name in the link
 			unless(defined($info)) {
 				if($cache) {
@@ -238,6 +257,10 @@ END {
 		$status = $1;
 	}
 
+	if($logger) {
+		$logger->debug("Initial status = $status");
+	}
+
 	# Generate the eTag before compressing, since the compressed data
 	# includes the mtime field which changes thus causing a different
 	# Etag to be generated
@@ -249,12 +272,11 @@ END {
 		require Encode;
 		$encode_loaded = 1;
 		$etag = '"' . Digest::MD5->new->add(Encode::encode_utf8($body))->hexdigest() . '"';
-		push @o, "ETag: $etag";
 		if($ENV{'HTTP_IF_NONE_MATCH'} && $generate_304 && ($status == 200)) {
 			if($logger) {
 				$logger->debug("Compare $ENV{HTTP_IF_NONE_MATCH} with $etag");
 			}
-			if($etag =~ /\Q$ENV{'HTTP_IF_NONE_MATCH'}\E/) {
+			if($ENV{'HTTP_IF_NONE_MATCH'} eq $etag) {
 				push @o, "Status: 304 Not Modified";
 				$send_body = 0;
 				$status = 304;
@@ -270,6 +292,7 @@ END {
 
 	if((length($encoding) > 0) && defined($body)) {
 		my $range = $ENV{'Range'} ? $ENV{'Range'} : $ENV{'HTTP_RANGE'};
+
 		if($range && !$cache) {
 			# TODO: Partials
 			if($range =~ /^bytes=(\d*)-(\d*)/) {
@@ -315,6 +338,14 @@ END {
 					$cache_hash = Storable::thaw($cobject->value());
 					$headers = $cache_hash->{'headers'};
 					@o = ("X-CGI-Buffer-$VERSION: Hit");
+					if($info) {
+						my $host_name = $info->host_name();
+						push @o, "X-Cache: HIT from $host_name";
+						push @o, "X-Cache-Lookup: HIT from $host_name";
+					} else {
+						push @o, 'X-Cache: HIT';
+						push @o, 'X-Cache-Lookup: HIT';
+					}
 				} else {
 					carp "Error retrieving data for key $key";
 				}
@@ -342,9 +373,13 @@ END {
 					foreach my $k (keys %{$cache_hash}) {
 						$body .= "\t$k\n";
 					}
-					warn($body);
 					$cache->remove($key);
-					carp "Can't retrieve body for key $key";
+					if($logger) {
+						$logger->error("Can't retrieve body for key $key");
+					} else {
+						carp "Can't retrieve body for key $key";
+					}
+					warn($body);
 					$send_body = 0;
 					$status = 500;
 				}
@@ -360,7 +395,10 @@ END {
 						}
 						$etag = '"' . Digest::MD5->new->add(Encode::encode_utf8($body))->hexdigest() . '"';
 					}
-					if(($etag =~ /\Q$ENV{'HTTP_IF_NONE_MATCH'}\E/) && $generate_304) {
+					if($logger && $generate_304) {
+						$logger->debug("Compare etags $ENV{HTTP_IF_NONE_MATCH} and $etag");
+					}
+					if(($ENV{'HTTP_IF_NONE_MATCH'} eq $etag) && $generate_304) {
 						push @o, "Status: 304 Not Modified";
 						$status = 304;
 						$send_body = 0;
@@ -379,6 +417,9 @@ END {
 							require Encode;
 							$encode_loaded = 1;
 						}
+						if($generate_etag && !defined($etag) && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
+							$etag = '"' . Digest::MD5->new->add(Encode::encode_utf8($body))->hexdigest() . '"';
+						}
 						my $nbody = Compress::Zlib::memGzip(\Encode::encode_utf8($body));
 						if(length($nbody) < length($body)) {
 							$body = $nbody;
@@ -389,13 +430,15 @@ END {
 				}
 			}
 			my $cannot_304 = !$generate_304;
-			if(defined($headers) && ($headers =~ /^ETag: "([a-z0-9]{32})"/m)) {
-				$etag = $1;
-			} else {
-				$etag = $cache_hash->{'etag'};
+			unless($etag) {
+				if(defined($headers) && ($headers =~ /^ETag: "([a-z0-9]{32})"/m)) {
+					$etag = $1;
+				} else {
+					$etag = $cache_hash->{'etag'};
+				}
 			}
 			if($ENV{'HTTP_IF_NONE_MATCH'} && $send_body && ($status != 304) && $generate_304) {
-				if(defined($etag) && ($etag =~ /\Q$ENV{'HTTP_IF_NONE_MATCH'}\E/) && ($status == 200)) {
+				if(defined($etag) && ($etag eq $ENV{'HTTP_IF_NONE_MATCH'}) && ($status == 200)) {
 					push @o, "Status: 304 Not Modified";
 					$send_body = 0;
 					$status = 304;
@@ -405,16 +448,18 @@ END {
 				} else {
 					$cannot_304 = 1;
 				}
-			} elsif($generate_etag && defined($etag) && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
-				push @o, "ETag: $etag";
 			}
 			if($cobject) {
-				if($ENV{'HTTP_IF_MODIFIED_SINCE'} && ($status != 304) && (!$cannot_304)) {
+				if($ENV{'HTTP_IF_MODIFIED_SINCE'} && ($status != 304) && !$cannot_304) {
 					_check_modified_since({
 						since => $ENV{'HTTP_IF_MODIFIED_SINCE'},
 						modified => $cobject->created_at()
 					});
-				} elsif($generate_last_modified) {
+				}
+				if(($status == 200) && $generate_last_modified) {
+					if($logger) {
+						$logger->debug('Set Last-Modified to ' . HTTP::Date::time2str($cobject->created_at()));
+					}
 					push @o, "Last-Modified: " . HTTP::Date::time2str($cobject->created_at());
 				}
 			}
@@ -427,11 +472,7 @@ END {
 					$cache_age = '10 minutes';
 				}
 				$cache_hash->{'body'} = $unzipped_body;
-				if($body && $send_body) {
-					my $body_length = length($body);
-					push @o, "Content-Length: $body_length";
-				}
-				if(scalar(@o)) {
+				if(@o && scalar(@o)) {
 					# Remember, we're storing the UNzipped
 					# version in the cache
 					my $c;
@@ -470,14 +511,53 @@ END {
 					}
 				}
 			}
+			if($info) {
+				my $host_name = $info->host_name();
+				if(defined($x_cache)) {
+					push @o, "X-Cache: $x_cache from $host_name";
+				} else {
+					push @o, "X-Cache: MISS from $host_name";
+				}
+				push @o, "X-Cache-Lookup: MISS from $host_name";
+			} else {
+				if(defined($x_cache)) {
+					push @o, "X-Cache: $x_cache";
+				} else {
+					push @o, 'X-Cache: MISS';
+				}
+				push @o, 'X-Cache-Lookup: MISS';
+			}
 			push @o, "X-CGI-Buffer-$VERSION: Miss";
 		}
 		# We don't need it any more, so give Perl a chance to
 		# tidy it up seeing as we're in the destructor
 		$cache = undef;
+	} elsif($info) {
+		my $host_name = $info->host_name();
+		push @o, ("X-Cache: MISS from $host_name", "X-Cache-Lookup: MISS from $host_name");
+	} else {
+		push @o, ('X-Cache: MISS', 'X-Cache-Lookup: MISS');
+	}
+	if($generate_etag && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
+		if(defined($etag)) {
+			push @o, "ETag: $etag";
+			if($logger) {
+				$logger->debug("Set ETag to $etag");
+			}
+		} elsif($logger && (($status == 200) || $status == 304) && !is_cached()) {
+			$logger->warn("BUG: ETag not generated, status $status");
+		}
 	}
 
-	my $body_length = defined($body) ? length($body) : 0;
+	my $body_length;
+	if(defined($body)) {
+		if(utf8::is_utf8($body)) {
+			utf8::encode($body);
+		}
+		$body_length = length($body);
+	} else {
+		$body_length = 0;
+	}
 
 	if(defined($headers) && length($headers)) {
 		# Put the original headers first, then those generated within
@@ -500,8 +580,7 @@ END {
 	}
 
 	if($body_length && $send_body) {
-		push @o, '';
-		push @o, $body;
+		push @o, ('', $body);
 	}
 
 	# XXXXXXXXXXXXXXXXXXXXXXX
@@ -512,6 +591,14 @@ END {
 		$SIG{__WARN__} = sub {
 			$wideCharWarningsIssued += "@_" =~ /Wide character in .../;
 			$widemess = "@_";
+			if($logger) {
+				$logger->fatal($widemess);
+				my $i = 1;
+				$logger->trace('Stack Trace');
+				while((my @call_details = (caller($i++)))) {
+					$logger->trace($call_details[1] . ':' . $call_details[2] . ' in function ' . $call_details[3]);
+				}
+			}
 			CORE::warn(@_);     # call the builtin warn as usual
 		};
 
@@ -538,6 +625,10 @@ END {
 }
 
 sub _check_modified_since {
+	if($logger) {
+		$logger->trace('In _check_modified_since');
+	}
+
 	if(!$generate_304) {
 		return;
 	}
@@ -596,7 +687,7 @@ sub _optimise_content {
 	$body =~ s/\n\s+|\s+\n/\n/g;
 	$body =~ s/\t+/ /g;
 	$body =~ s/\s(\<.+?\>\s\<.+?\>)/$1/;
-	$body =~ s/(\<.+?\>\s\<.+?\>)\s/$1/g;
+	# $body =~ s/(\<.+?\>\s\<.+?\>)\s/$1/g;
 	$body =~ s/\<p\>\s/\<p\>/gi;
 	$body =~ s/\<\/p\>\s\<p\>/\<\/p\>\<p\>/gi;
 	$body =~ s/\<\/tr\>\s\<tr\>/\<\/tr\>\<tr\>/gi;
@@ -660,6 +751,7 @@ Set various options and override default values.
 	optimise_content => 0,	# optimise your program's HTML, CSS and JavaScript
 	cache => CHI->new(driver => 'File'),	# cache requests
 	cache_key => 'string',	# key for the cache
+	cache_age => '10 minutes',	# how long to store responses in the cache
 	logger => $logger,
 	lint->content => 0,	# Pass through HTML::Lint
 	generate_304 => 1,	# Generate 304: Not modified
@@ -670,7 +762,12 @@ The cache_key should be a unique value dependent upon the values set by the
 browser.
 
 The cache object will be an object that understands get_object(),
-set(), remove() and created_at() messages, such as an L<CHI> object.
+set(), remove() and created_at() messages, such as an L<CHI> object. It is
+used as a server-side cache to reduce the need to rerun database accesses.
+
+Items stay in the server-side cache by default for 10 minutes.
+This can be overridden by the cache_control HTTP header in the request, and
+the default can be changed by the cache_age argument to init().
 
 Logger will be an object that understands debug() such as an L<Log::Log4perl>
 object.
@@ -745,6 +842,7 @@ sub init {
 				}
 			}
 		}
+		$cache_age ||= $params{cache_age};
 		$cache = $params{cache};
 		if(defined($params{cache_key})) {
 			$cache_key = $params{cache_key};
@@ -768,9 +866,9 @@ Synonym for init, kept for historical reasons.
 =cut
 
 sub set_options {
-	my %params = @_;
+	my %params = (ref($_[0]) eq 'HASH') ? %{$_[0]} : @_;
 
-	init(%params);
+	init(\%params);
 }
 
 =head2 can_cache
@@ -780,7 +878,12 @@ Returns true if the server is allowed to store the results locally.
 =cut
 
 sub can_cache {
+	if(defined($x_cache)) {
+		return ($x_cache eq 'HIT');
+	}
+
 	if(defined($ENV{'NO_CACHE'}) || defined($ENV{'NO_STORE'})) {
+		$x_cache = 'MISS';
 		return 0;
 	}
 	if(defined($ENV{'HTTP_CACHE_CONTROL'})) {
@@ -790,10 +893,13 @@ sub can_cache {
 		}
 		if(($control eq 'no-store') ||
 		       ($control eq 'no-cache') ||
+		       ($control eq 'max-age=0') ||
 		       ($control eq 'private')) {
+			$x_cache = 'MISS';
 			return 0;
 		}
 	}
+	$x_cache = 'HIT';
 	return 1;
 }
 
@@ -883,7 +989,8 @@ sub is_cached {
 		$cobject = undef;
 		# Nothing will be in date and all new searches would miss
 		# anyway, so may as well clear it all
-		$cache->clear();
+		# FIXME: RT104471
+		# $cache->clear();
 		return 0;
 	}
 	if($logger) {
@@ -915,11 +1022,11 @@ sub _my_age {
 }
 
 sub _should_gzip {
-	if($compress_content && ($ENV{'HTTP_ACCEPT_ENCODING'} || $ENV{'HTTP_TE'})) {
+	if($compress_content && $send_body && ($ENV{'HTTP_ACCEPT_ENCODING'} || $ENV{'HTTP_TE'})) {
 		my $accept = lc($ENV{'HTTP_ACCEPT_ENCODING'} ? $ENV{'HTTP_ACCEPT_ENCODING'} : $ENV{'HTTP_TE'});
 		foreach my $encoding ('x-gzip', 'gzip') {
 			$_ = $accept;
-			if($content_type[0]) {
+			if(@content_type && $content_type[0]) {
 				if (m/$encoding/i && (lc($content_type[0]) eq 'text')) {
 					return $encoding;
 				}
@@ -976,6 +1083,8 @@ L<Log::Log4Perl>, if you're using that, so that any messages it
 produces are printed after the HTTP headers have been sent by
 CGI::Buffer;
 
+CGI::Buffer is not compatible with FCGI, use L<FCGI::Buffer> instead.
+
 Please report any bugs or feature requests to C<bug-cgi-buffer at rt.cpan.org>,
 or through the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=CGI-Buffer>.
 I will be notified, and then you'll automatically be notified of progress on
@@ -1030,7 +1139,7 @@ The licence for cgi_buffer is:
 
     This software is provided 'as is' without warranty of any kind."
 
-The rest of the program is Copyright 2011-2014 Nigel Horne,
+The rest of the program is Copyright 2011-2015 Nigel Horne,
 and is released under the following licence: GPL
 
 =cut
